import os
import sys

import torch
from tfkit.utility.predictor import ClassificationPredictor
from torch import nn

dir_path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.abspath(os.path.join(dir_path, os.pardir)))

from torch import softmax, sigmoid
from tfkit.task.clas import Preprocessor
from tfkit.utility.loss import FocalLoss, BCEFocalLoss


class Model(nn.Module):

    def __init__(self, tokenizer, pretrained, tasks_detail, maxlen=512, dropout=0.1, **kwargs):
        super().__init__()
        self.tokenizer = tokenizer
        self.pretrained = pretrained

        self.dropout = nn.Dropout(dropout)
        self.loss_fct = FocalLoss()
        self.loss_fct_mt = BCEFocalLoss()

        self.tasks = dict()
        self.tasks_detail = tasks_detail
        self.classifier_list = nn.ModuleList()
        for task, labels in tasks_detail.items():
            self.classifier_list.append(nn.Linear(self.pretrained.config.hidden_size, len(labels)))
            self.tasks[task] = len(self.classifier_list) - 1
        self.maxlen = maxlen

        self.pretrained = self.pretrained
        self.classifier_list = self.classifier_list
        self.loss_fct = self.loss_fct
        self.loss_fct_mt = self.loss_fct_mt

        predictor = ClassificationPredictor(self, Preprocessor)
        self.predictor = predictor
        self.predict = predictor.predict

    def get_all_task(self):
        """
        list all classification task
        :return: tasks list
        """
        return list(self.tasks.keys())

    def mean_pooling(self, model_output, attention_mask):
        """
        Mean Pooling - Take attention mask into account for correct averaging
        from https://github.com/UKPLab/sentence-transformers
        modify - mask from -1 to 0
        :param model_output:
        :param attention_mask:
        :return:
        """
        input_mask_expanded = attention_mask.unsqueeze(-1).expand(model_output.size()).float()
        input_mask_expanded[input_mask_expanded < 0] = 0
        sum_embeddings = torch.sum(model_output * input_mask_expanded, 1)
        sum_mask = torch.clamp(input_mask_expanded.sum(1), min=1e-9)
        return sum_embeddings / sum_mask

    def forward(self, batch_data, eval=False, **kwargs):
        # covert input to correct data type
        tasks = batch_data['task']
        tasks = [bytes(t).decode(encoding="utf-8", errors="ignore") for t in tasks]
        inputs = torch.as_tensor(batch_data['input'])
        targets = torch.as_tensor(batch_data['target'])
        masks = torch.as_tensor(batch_data['mask'])
        # define model output
        result_dict = {
            'max_item': [],
            'prob_list': [],
            'label_prob': []
        }

        result_logits = []
        result_labels = []
        for p, zin in enumerate(zip(tasks, inputs, masks)):
            task, input, mask = zin
            task_id = self.tasks[task]
            task_labels = self.tasks_detail[task]

            output = self.pretrained(input.unsqueeze(0), mask.unsqueeze(0))[0]
            pooled_output = self.dropout(self.mean_pooling(output, mask.unsqueeze(0)))
            classifier_output = self.classifier_list[task_id](pooled_output)
            reshaped_logit = classifier_output.view(-1, len(task_labels))  # 0 for cls position
            result_logits.append(reshaped_logit)
            if not eval:
                target = targets[p]
                result_labels.append(target)
            else:
                if 'multi_label' in task:
                    reshaped_logit = sigmoid(reshaped_logit)
                else:
                    reshaped_logit = softmax(reshaped_logit, dim=1)
                logit_prob = reshaped_logit[0].data.tolist()
                logit_label = dict(zip(task_labels, logit_prob))
                result_dict['label_prob'].append({task: logit_label})
                if 'multi_label' in task:
                    result_dict['max_item'].append({task: [k for k, v in logit_label.items() if v > 0.5]})
                else:
                    result_dict['max_item'].append({task: [task_labels[logit_prob.index(max(logit_prob))]]})

        if eval:
            outputs = result_dict
        else:
            loss = 0
            for logit, labels, task in zip(result_logits, result_labels, tasks):
                if 'multi_label' in task:
                    loss += self.loss_fct_mt(logit, labels.type_as(logit))
                else:
                    loss += self.loss_fct(logit, labels)
            outputs = loss

        return outputs



import unittest

import tfkit
from tfkit.test import *


class TestEval(unittest.TestCase):

    def testHelp(self):
        result = os.system('tfkit-eval -h')
        self.assertTrue(result == 0)

    def test_parser(self):
        parser, _ = tfkit.eval.parse_eval_args(
            ['--model', 'once', '--metric', 'emf1', '--valid', 'test.csv', '--print'])
        print(parser)
        self.assertTrue(parser.get('model') == ['once'])

        eval_parser, model_parser = tfkit.eval.parse_eval_args(
            ['--model', 'once', '--metric', 'emf1', '--valid', 'test.csv', '--print', '--decodenum', '2'])
        self.assertTrue(eval_parser.get('model') == ['once'])
        self.assertTrue(model_parser.get('decodenum') == '2')

    def testEvalGen(self):
        tfkit.eval.main(
            ['--model', ONCE_MODEL_PATH, '--valid', GEN_DATASET, '--metric', 'emf1', '--print'])
        result = os.system(
            'tfkit-eval --model ' + ONCE_MODEL_PATH + ' --valid ' + GEN_DATASET + ' --metric emf1 --print')
        self.assertTrue(result == 0)

    def testEvalGenOnce(self):
        tfkit.eval.main(
            ['--model', ONCE_MODEL_PATH, '--valid', GEN_DATASET, '--metric', 'emf1', '--print'])
        result = os.system(
            'tfkit-eval --model ' + ONCE_MODEL_PATH + ' --valid ' + GEN_DATASET + ' --metric emf1 --print')
        self.assertTrue(result == 0)

    def testEvalGenOnceCTC(self):
        tfkit.eval.main(
            ['--model', ONCECTC_MODEL_PATH, '--valid', GEN_DATASET, '--metric', 'emf1', '--print'])
        result = os.system(
            'tfkit-eval --model ' + ONCECTC_MODEL_PATH + ' --valid ' + GEN_DATASET + ' --metric emf1 --print')
        self.assertTrue(result == 0)

    def testEvalSeq2Seq(self):
        tfkit.eval.main(
            ['--model', SEQ2SEQ_MODEL_PATH, '--valid', GEN_DATASET, '--metric', 'emf1', '--print',
             '--decodenum', '2'])
        tfkit.eval.main(
            ['--model', SEQ2SEQ_MODEL_PATH, '--valid', GEN_DATASET, '--metric', 'emf1', '--print'])
        result = os.system(
            'tfkit-eval --model ' + SEQ2SEQ_MODEL_PATH + ' --valid ' + GEN_DATASET + ' --metric emf1 --print')
        self.assertTrue(result == 0)

    def testEvalCLM(self):
        tfkit.eval.main(
            ['--model', CLM_MODEL_PATH, '--valid', GEN_DATASET, '--metric', 'emf1', '--print'])
        result = os.system(
            'tfkit-eval --model ' + CLM_MODEL_PATH + ' --valid ' + GEN_DATASET + ' --metric emf1 --print')
        self.assertTrue(result == 0)

    def testEvalAddedTokenModel(self):
        result = os.system(
            'tfkit-eval --model ' + ADDTOKFILE_MODEL_PATH + ' --valid ' + ADDTOK_DATASET + ' --metric emf1 --print')
        self.assertTrue(result == 0)
        result = os.system(
            'tfkit-eval --model ' + ADDTOKFILE_MODEL_PATH + ' --valid ' + ADDTOK_DATASET + ' --metric emf1 --print')
        self.assertTrue(result == 0)
        result = os.system(
            'tfkit-eval --model ' + ADDTOKFREQ_MODEL_PATH + ' --valid ' + ADDTOK_DATASET + ' --metric emf1 --print')
        self.assertTrue(result == 0)
        result = os.system(
            'tfkit-eval --model ' + ADDTOKFREQ_MODEL_PATH + ' --valid ' + ADDTOK_DATASET + ' --metric emf1 --print')
        self.assertTrue(result == 0)

    # def testEvalQA(self):
    #     tfkit.eval.main(
    #         ['--model', QA_MODEL_PATH, '--valid', QA_DATASET, '--metric', 'emf1', '--print'])
    #     result = os.system(
    #         'tfkit-eval --model ' + QA_MODEL_PATH + ' --valid ' + QA_DATASET + ' --metric emf1 --print')
    #     self.assertTrue(result == 0)
    #
    # def testEvalClassify(self):
    #     tfkit.eval.main(
    #         ['--model', CLAS_MODEL_PATH, '--valid', CLAS_DATASET, '--metric', 'clas', '--print'])
    #     result = os.system(
    #         'tfkit-eval --model ' + CLAS_MODEL_PATH + ' --valid ' + CLAS_DATASET + ' --metric clas --print')
    #     self.assertTrue(result == 0)
    #
    # def testEvalTag(self):
    #     tfkit.eval.main(
    #         ['--model', TAG_MODEL_PATH, '--valid', TAG_DATASET, '--metric', 'clas', '--print'])
    #     result = os.system(
    #         'tfkit-eval --model ' + TAG_MODEL_PATH + ' --valid ' + TAG_DATASET + ' --metric clas --print')
    #     self.assertTrue(result == 0)


import numpy
import torch
from torch import nn
from torch.utils import data


def index_of(in_list, val):
    """
    get token index in list, return -1 when it is not in the list 
    :rtype: int
    :param in_list: query list
    :param val: query target
    :return: position index
    """
    try:
        return in_list.index(val)
    except ValueError:
        return -1


def pad_batch(batch):
    """
    reduce batch data shape by reduce their padding to common max
    it needs to Handel some exception since some key is no need to be padded
    :param batch: list of dict, with key input and target as model input and target
    :return: list of dict
    """
    keys = list(batch[0].keys())
    for k in keys:
        batch_key_length = [len(i[k]) if not isinstance(i[k], int) else 1 for i in batch]
        if len(set(batch_key_length)) > 1:  # is all value same? if no, it need to pad with max length
            pad_length = max(batch_key_length)
            for idx, _ in enumerate(batch):
                if f"{k}_pad" in batch[idx]:
                    padded = nn.ConstantPad1d((0, pad_length - len(batch[idx][k])), batch[idx][f"{k}_pad"][0])

                else:
                    padded = nn.ConstantPad1d((0, pad_length - len(batch[idx][k])), 0)
                # batch[idx][k] = torch.unsqueeze(padded(batch[idx][k]), 0)
                batch[idx][k] = padded(batch[idx][k])
    for ind, dat in enumerate(batch):
        for k, v in dat.items():
            batch[ind][k] = numpy.asarray(batch[ind][k])
    return batch


def dataloader_collate(batch):
    """
    dataloader_collate function to apply batch reduce padding
    :param batch: list of dict
    :return: batch: list of dict
    """
    # batch = copy.deepcopy(batch)
    return torch.utils.data._utils.collate.default_collate(pad_batch(batch))



import os
import sys

from tfkit.task.clm import Preprocessor
from tfkit.utility.predictor import AutoRegressivePredictor

dir_path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.abspath(os.path.join(dir_path, os.pardir)))

import torch
from torch import nn
from torch.nn.functional import softmax


class Model(nn.Module):
    def __init__(self, tokenizer, pretrained, maxlen=512, **kwargs):
        super().__init__()
        self.tokenizer = tokenizer
        self.pretrained = pretrained
        self.vocab_size = max(self.pretrained.config.vocab_size, self.tokenizer.__len__())
        self.model = nn.Linear(self.pretrained.config.hidden_size, self.vocab_size)
        self.maxlen = maxlen
        predictor = AutoRegressivePredictor(self, Preprocessor)
        self.predictor = predictor
        self.predict = predictor.predict

    def clean_cache(self):
        self.encoder_outputs = None
        self.past_key_values = None

    def forward(self, batch_data, eval=False, beamsearch=False, max_return=1, **kwargs):
        inputs = batch_data['input']
        masks = batch_data['mask']
        tokens_tensor = torch.as_tensor(inputs)
        mask_tensors = torch.as_tensor(masks)

        outputs = self.pretrained(tokens_tensor, attention_mask=mask_tensors)
        prediction_scores = self.model(outputs[0])

        if eval:
            result_dict = {}
            start = batch_data['start'][0]
            softmax_score = softmax(prediction_scores[0][start], dim=-1).flatten()
            max_item_id = torch.argmax(softmax_score, -1).item()
            max_item_prob = softmax_score[max_item_id].item()
            result_dict['max_item'] = (self.tokenizer.convert_ids_to_tokens(max_item_id), max_item_prob)
            if max_return > 1:
                topK = torch.topk(softmax_score, max_return)
                prob_result = [(self.tokenizer.convert_ids_to_tokens(tid), prob) for prob, tid in
                               zip(topK.values.data.tolist(), topK.indices.data.tolist())]
                result_dict['prob_list'] = softmax_score.data.tolist()[:max_return]
                result_dict['label_prob'] = prob_result
            outputs = result_dict
        else:
            targets = batch_data['target']
            loss_tensors = torch.as_tensor(targets)
            loss_fct = nn.CrossEntropyLoss(ignore_index=-1)  # -1 index = padding token
            masked_lm_loss = loss_fct(prediction_scores.view(-1, self.vocab_size),
                                      loss_tensors.view(-1))

            outputs = masked_lm_loss
        return outputs



import os

ROOT_DIR = os.path.dirname(os.path.abspath(__file__ + "/../../"))

DATASET_DIR = os.path.join(ROOT_DIR, 'demo_data')
TAG_DATASET = os.path.join(DATASET_DIR, 'tag.csv')
CLAS_DATASET = os.path.join(DATASET_DIR, 'classification.csv')
GEN_DATASET = os.path.join(DATASET_DIR, 'generation.csv')
MASK_DATASET = os.path.join(DATASET_DIR, 'mask.csv')
MCQ_DATASET = os.path.join(DATASET_DIR, 'mcq.csv')
QA_DATASET = os.path.join(DATASET_DIR, 'qa.csv')
ADDTOK_DATASET = os.path.join(DATASET_DIR, 'unk_tok.csv')
NEWTOKEN_FILE = os.path.join(DATASET_DIR, 'tok_list.txt')

MODEL_SAVE_DIR = os.path.join(ROOT_DIR, 'tfkit/test/cache/')
ADDTOKFREQ_SAVE_DIR = os.path.join(MODEL_SAVE_DIR, 'addtokfreq/')
ADDTOKFILE_SAVE_DIR = os.path.join(MODEL_SAVE_DIR, 'addtokfile/')
CLAS_MODEL_DIR = os.path.join(MODEL_SAVE_DIR, 'clas/')
TAG_MODEL_DIR = os.path.join(MODEL_SAVE_DIR, 'tag/')
TAGCRF_MODEL_DIR = os.path.join(MODEL_SAVE_DIR, 'tagcrf/')
ONEBYONE_MODEL_DIR = os.path.join(MODEL_SAVE_DIR, 'onebyone/')
CLM_MODEL_DIR = os.path.join(MODEL_SAVE_DIR, 'clm/')
SEQ2SEQ_MODEL_DIR = os.path.join(MODEL_SAVE_DIR, 'seq2seq/')
ONCE_MODEL_DIR = os.path.join(MODEL_SAVE_DIR, 'once/')
ONCECTC_MODEL_DIR = os.path.join(MODEL_SAVE_DIR, 'oncectc/')
MASK_MODEL_DIR = os.path.join(MODEL_SAVE_DIR, 'mask/')
MCQ_MODEL_DIR = os.path.join(MODEL_SAVE_DIR, 'mcq/')
QA_MODEL_DIR = os.path.join(MODEL_SAVE_DIR, 'qa/')
MTTASK_MODEL_DIR = os.path.join(MODEL_SAVE_DIR, 'mttask/')

ONEBYONE_MODEL_PATH = os.path.join(ONEBYONE_MODEL_DIR, '2.pt')
ONCE_MODEL_PATH = os.path.join(ONCE_MODEL_DIR, '2.pt')
ONCECTC_MODEL_PATH = os.path.join(ONCECTC_MODEL_DIR, '1.pt')
SEQ2SEQ_MODEL_PATH = os.path.join(SEQ2SEQ_MODEL_DIR, '2.pt')
CLM_MODEL_PATH = os.path.join(CLM_MODEL_DIR, '2.pt')
CLAS_MODEL_PATH = os.path.join(CLAS_MODEL_DIR, '2.pt')
MASK_MODEL_PATH = os.path.join(MASK_MODEL_DIR, '2.pt')
MCQ_MODEL_PATH = os.path.join(MCQ_MODEL_DIR, '2.pt')
TAG_MODEL_PATH = os.path.join(TAG_MODEL_DIR, '2.pt')
QA_MODEL_PATH = os.path.join(QA_MODEL_DIR, '2.pt')
ADDTOKFREQ_MODEL_PATH = os.path.join(ADDTOKFREQ_SAVE_DIR, '2.pt')
ADDTOKFILE_MODEL_PATH = os.path.join(ADDTOKFILE_SAVE_DIR, '2.pt')



import os
import sys

from tfkit.utility.predictor import QuestionAnsweringPredictor

dir_path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.abspath(os.path.join(dir_path, os.pardir)))

import torch
import torch.nn as nn
from torch.nn.functional import softmax
from tfkit.task.qa.preprocessor import Preprocessor


class Model(nn.Module):

    def __init__(self, tokenizer, pretrained, maxlen=128, dropout=0.1, **kwargs):
        super().__init__()
        self.tokenizer = tokenizer
        self.pretrained = pretrained
        self.maxlen = maxlen

        self.dropout = nn.Dropout(dropout)
        self.loss_fct = nn.CrossEntropyLoss(ignore_index=-1)
        # self.loss_fct = FocalLoss(ignore_index=-1)
        # self.loss_fct = GWLoss()

        self.pretrained = self.pretrained
        self.qa_classifier = nn.Linear(self.pretrained.config.hidden_size, 2)
        self.loss_fct = self.loss_fct

        predictor = QuestionAnsweringPredictor(self, Preprocessor)
        self.predictor = predictor
        self.predict = predictor.predict

    def forward(self, batch_data, eval=False, **kwargs):
        print("batch_data",batch_data)
        inputs = torch.as_tensor(batch_data['input'])
        masks = torch.as_tensor(batch_data['mask'])
        targets = torch.as_tensor(batch_data['target'])
        start_positions, end_positions = targets.split(1, dim=1)
        start_positions = start_positions.squeeze(1)
        end_positions = end_positions.squeeze(1)

        output = self.pretrained(inputs, attention_mask=masks)[0]
        logits = self.qa_classifier(output)
        start_logits, end_logits = logits.split(1, dim=-1)
        start_logits = start_logits.squeeze(-1)
        end_logits = end_logits.squeeze(-1)

        if eval:
            result_dict = {
                'label_prob_all': [],
                'label_map': []
            }
            reshaped_start_logits = softmax(start_logits, dim=1)
            reshaped_end_logits = softmax(end_logits, dim=1)
            start_prob = reshaped_start_logits.data.tolist()[0]
            end_prob = reshaped_end_logits.data.tolist()[0]
            result_dict['label_prob_all'].append({'start': dict(zip(range(len(start_prob)), start_prob)),
                                                  'end': dict(zip(range(len(end_prob)), end_prob))})
            result_dict['label_map'].append({'start': start_prob.index(max(start_prob)),
                                             'end': end_prob.index(max(end_prob))})
            outputs = result_dict
        else:
            start_loss = self.loss_fct(start_logits, start_positions)
            end_loss = self.loss_fct(end_logits, end_positions)
            total_loss = (start_loss + end_loss) / 2
            outputs = total_loss

        return outputs



import os
import sys

from transformers import AutoModel

from tfkit.task.seq2seq import Preprocessor
from tfkit.utility.model import tie_encoder_decoder_weights
from tfkit.utility.predictor import AutoRegressivePredictor

dir_path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.abspath(os.path.join(dir_path, os.pardir)))

from torch.autograd.grad_mode import F
import torch
from torch import nn
from torch.nn.functional import softmax
from tfkit.utility.loss import NegativeCElLoss, SelfKDLoss
import copy


class Model(nn.Module):
    def __init__(self, tokenizer, pretrained, maxlen=512, selfkd=False, **kwargs):
        super().__init__()
        self.maxlen = maxlen
        self.tokenizer = tokenizer
        self.pretrained = pretrained
        self.selfkd = selfkd
        init_weight = None
        if hasattr(pretrained, 'decoder'):
            self.decoder_model = None
            decoder_hidden_size = pretrained.config.hidden_size
            if hasattr(pretrained, 'shared'):
                init_weight = copy.deepcopy(pretrained.shared.weight)
        else:
            decoder_config = copy.deepcopy(pretrained.config)
            decoder_config.is_decoder = True
            decoder_config.add_cross_attention = True
            self.decoder_model = AutoModel.from_config(decoder_config)
            tie_encoder_decoder_weights(
                self.pretrained, self.decoder_model,
                self.decoder_model.base_model_prefix
            )
            decoder_hidden_size = decoder_config.hidden_size

        self.vocab_size = max(self.pretrained.config.vocab_size, self.tokenizer.__len__())
        self.model = nn.Linear(decoder_hidden_size, self.vocab_size, bias=False)
        if init_weight is not None:
            self.model.weight = init_weight
        self.encoder_outputs = None
        self.past_key_values = None
        predictor = AutoRegressivePredictor(self, Preprocessor)
        self.predictor = predictor
        self.predict = predictor.predict

    def clean_cache(self):
        self.encoder_outputs = None
        self.past_key_values = None

    def forward(self, batch_data, eval=False, beamsearch=False, max_return=1, **kwargs):
        input_tensors = torch.as_tensor(batch_data['input'])
        prev_tensors = torch.as_tensor(batch_data['prev'])
        encoder_mask_tensors = torch.as_tensor(batch_data['encoder_mask'])
        decoder_mask_tensors = torch.as_tensor(batch_data['decoder_mask'])
        if self.decoder_model:
            if not eval or self.encoder_outputs is None:
                outputs = self.pretrained(input_tensors, attention_mask=encoder_mask_tensors)
                self.encoder_outputs = outputs[0]
            # Decoder
            prediction = self.decoder_model(
                input_ids=prev_tensors,
                attention_mask=decoder_mask_tensors,
                encoder_hidden_states=self.encoder_outputs,
                output_hidden_states=self.selfkd,
                use_cache=True,
                return_dict=True,
            )
            prediction_output = prediction['last_hidden_state']
            prediction_all_hidden = prediction.get('hidden_states')
        else:
            if eval and self.encoder_outputs is not None and not beamsearch:
                prev_tensors = prev_tensors[..., -1:]
                prediction = self.pretrained(
                    encoder_outputs=self.encoder_outputs,
                    decoder_input_ids=prev_tensors,
                    past_key_values=self.past_key_values,
                    output_hidden_states=self.selfkd,
                    use_cache=True,
                    return_dict=True
                )
            elif eval and self.encoder_outputs is not None:
                prev_tensors = prev_tensors[..., -1:]
                prediction = self.pretrained(
                    encoder_outputs=self.encoder_outputs,
                    decoder_input_ids=prev_tensors,
                    past_key_values=self.past_key_values,
                    output_hidden_states=self.selfkd,
                    use_cache=True,
                    return_dict=True
                )
            else:
                prediction = self.pretrained(
                    input_ids=input_tensors,
                    attention_mask=encoder_mask_tensors,
                    decoder_input_ids=prev_tensors,
                    decoder_attention_mask=decoder_mask_tensors,
                    past_key_values=self.past_key_values,
                    output_hidden_states=self.selfkd,
                    use_cache=True,
                    return_dict=True
                )
            prediction_output = prediction['last_hidden_state']
            prediction_all_hidden = prediction.get('decoder_hidden_states')
            if eval and self.encoder_outputs is None:
                self.encoder_outputs = (prediction['encoder_last_hidden_state'])
            if eval and not beamsearch:
                self.past_key_values = prediction.get('past_key_values', None)
        prediction_scores = self.model(prediction_output)
        if eval:
            result_dict = {}
            softmax_score = softmax(prediction_scores[0][0], dim=0)
            max_item_id = torch.argmax(softmax_score, -1).item()
            max_item_prob = softmax_score[max_item_id].item()
            result_dict['max_item'] = (self.tokenizer.convert_ids_to_tokens(max_item_id), max_item_prob)
            if max_return > 1:
                topK = torch.topk(softmax_score, max_return)
                prob_result = [(self.tokenizer.convert_ids_to_tokens(tid), prob) for prob, tid in
                               zip(topK.values.data.tolist(), topK.indices.data.tolist())]
                result_dict['prob_list'] = softmax_score.data.tolist()[:max_return]
                result_dict['label_prob'] = prob_result
            outputs = result_dict
        else:
            targets = batch_data['target']
            negative_targets = batch_data['ntarget']
            loss_tensors = torch.as_tensor(targets)
            loss_fct = nn.CrossEntropyLoss(ignore_index=-1)  # -1 index = padding token
            lm_loss = loss_fct(prediction_scores.view(-1, self.vocab_size),
                               loss_tensors.view(-1))
            if self.selfkd:
                selfkdloss_fct = SelfKDLoss(ignore_index=-1)
                for decoder_hidden in prediction_all_hidden[:-1]:
                    student = self.model(decoder_hidden)
                    lm_loss += selfkdloss_fct(student.view(-1, self.vocab_size),
                                              prediction_scores.view(-1, self.vocab_size), loss_tensors.view(-1))
            if 'btarget' in batch_data:
                backtran_tensors = torch.as_tensor(batch_data['btarget'])
                if not torch.all(backtran_tensors.eq(-1)).item():
                    backtran_predation = self.pretrained(
                        input_ids=backtran_tensors,
                        output_hidden_states=True,
                        return_dict=True
                    )
                    backtran_hidden = backtran_predation[
                        'encoder_last_hidden_state'] if 'encoder_last_hidden_state' in backtran_predation else \
                        prediction[
                            'last_hidden_state']
                    backtran_loss = F.cosine_similarity(self.encoder_hidden, backtran_hidden).mean()
                    lm_loss += backtran_loss
            negativeloss_tensors = torch.as_tensor(negative_targets)
            if not torch.all(negativeloss_tensors.eq(-1)).item():
                negative_loss_fct = NegativeCElLoss(ignore_index=-1)
                negative_loss = negative_loss_fct(prediction_scores.view(-1, self.vocab_size),
                                                  negativeloss_tensors.view(-1))
                lm_loss += negative_loss
            outputs = lm_loss
        return outputs



import csv
import sys
import os

from tfkit.utility.logger import Logger

dir_path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.abspath(os.path.join(dir_path, os.pardir)))

import unittest
import tfkit


class TestLogger(unittest.TestCase):
    ROOT_DIR = os.path.dirname(os.path.abspath(__file__ + "/../../"))
    MODEL_SAVE_PATH = os.path.join(ROOT_DIR, './test/cache/')

    def test_write_log(self):
        logger = Logger(savedir=self.MODEL_SAVE_PATH)
        logger.write_log("test")
        with open(logger.logfilepath, 'r') as f:
            lines = f.read().splitlines()
            last_line = lines[-1]
            print(last_line)
            self.assertEqual(last_line, "test")

    def test_write_metric(self):
        logger = Logger(savedir=self.MODEL_SAVE_PATH)
        logger.write_metric("test", 1, 0)
        with open(logger.metricfilepath, 'r') as f:
            last_row = list(csv.reader(f))[-1]
            self.assertEqual(last_row, ["test", '1', '0'])



import os
import sys

import numpy
from transformers import AutoTokenizer

from tfkit.utility.dataset import get_dataset, TFKitDataset
from tfkit.utility.model import load_pretrained_tokenizer, load_model_class

dir_path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.abspath(os.path.join(dir_path, os.pardir)))

import unittest
import tfkit
from tfkit.test import *


class TestDataset(unittest.TestCase):

    def check_type_for_dataloader(self, data_item):
        if (isinstance(data_item, list) and not isinstance(data_item[-1], str) and self.check_type_for_dataloader(
                data_item[-1])) or \
                isinstance(data_item, numpy.ndarray) or \
                isinstance(data_item, int):
            return True
        else:
            return False

    def test_check_type_for_dataloader(self):
        self.assertTrue(self.check_type_for_dataloader(123))
        self.assertFalse(self.check_type_for_dataloader('a'))
        self.assertFalse(self.check_type_for_dataloader(['a']))
        self.assertFalse(self.check_type_for_dataloader([{'a'}]))

    def test_get_dataset(self):
        tokenizer = load_pretrained_tokenizer('voidful/albert_chinese_tiny')
        model_class = load_model_class('clas')
        file_path = CLAS_DATASET
        dataset_arg = {
            'maxlen': 123,
            'handle_exceed': 'slide',
            'config': 'voidful/albert_chinese_tiny',
            'cache': False
        }
        ds = get_dataset(file_path=file_path, task_class=model_class, tokenizer=tokenizer,
                                       parameter=dataset_arg)
        print(ds, ds[0])

    # def testClassifier(self):
    #     for i in tfkit.task.clas.get_data_from_file(CLAS_DATASET):
    #         print(i)
    #     tokenizer = AutoTokenizer.from_pretrained('voidful/albert_chinese_tiny')
    #
    #     for i in TFKitDataset(CLAS_DATASET,
    #                          tokenizer=tokenizer,
    #                          get_data_from_file=tfkit.task.clas.get_data_from_file,
    #                          preprocessor=tfkit.task.clas.preprocessor,
    #                          get_feature_from_data=tfkit.task.clas.get_feature_from_data):
    #         print(i)
    #         self.assertTrue(len(i['input']) <= 512)
    #         self.assertTrue(len(i['target']) <= 512)
    # def testProcessingArg_maxlen(self):
    #     maxlen = 128
    #     tokenizer = AutoTokenizer.from_pretrained('voidful/albert_chinese_small')
    #     for m in [tfkit.task.seq2seq, tfkit.task.clas]:
    #         for i in TFKitDataset(GEN_DATASET,
    #                              tokenizer=tokenizer,
    #                              get_data_from_file=m.get_data_from_file,
    #                              preprocessor=m.preprocessor,
    #                              get_feature_from_data=m.get_feature_from_data,
    #                              preprocessing_arg={'maxlen': maxlen, }):
    #             print(tokenizer.convert_tokens_to_string(tokenizer.convert_ids_to_tokens(i['input'])))
    #             print(len(i['input']), i['input'])
    #             self.assertTrue(len(i['input']) <= maxlen)
    #             print(len(i['target']), i['target'])
    #             self.assertTrue(len(i['target']) <= maxlen)
    #
    # def testSeq2seq(self):
    #     tokenizer = AutoTokenizer.from_pretrained('facebook/bart-base')
    #
    #     for i in tfkit.task.seq2seq.get_data_from_file(GEN_DATASET):
    #         print(i)
    #     for i in TFKitDataset(GEN_DATASET,
    #                          tokenizer=tokenizer,
    #                          get_data_from_file=tfkit.task.seq2seq.get_data_from_file,
    #                          preprocessor=tfkit.task.seq2seq.preprocessor,
    #                          get_feature_from_data=tfkit.task.seq2seq.get_feature_from_data):
    #         print(i.keys())
    #         self.assertTrue(len(i['prev']) == 512)
    #         self.assertTrue(len(i['input']) == 512)
    #         self.assertTrue(len(i['target']) == 512)
    #
    #     maxlen = 100
    #     for likelihood in ['none', 'neg', 'pos', 'both']:
    #         for i in TFKitDataset(GEN_DATASET,
    #                              tokenizer=tokenizer,
    #                              get_data_from_file=tfkit.task.seq2seq.get_data_from_file,
    #                              preprocessor=tfkit.task.seq2seq.preprocessor,
    #                              get_feature_from_data=tfkit.task.seq2seq.get_feature_from_data,
    #                              preprocessing_arg={'maxlen': maxlen, 'likelihood': likelihood}):
    #             print(likelihood, i)
    #             print(tokenizer.convert_tokens_to_string(tokenizer.convert_ids_to_tokens(i['input'])))
    #             print(len(i['input']), i['input'])
    #             self.assertTrue(len(i['input']) == maxlen)
    #             print(len(i['target']), i['target'])
    #             self.assertTrue(len(i['target']) == maxlen)
    #
    # def testTag(self):
    #     tokenizer = AutoTokenizer.from_pretrained('voidful/albert_chinese_tiny')
    #     for i in tfkit.tag.get_data_from_file(TAG_DATASET):
    #         print(i)
    #     maxlen = 128
    #     for i in TFKitDataset(TAG_DATASET,
    #                          tokenizer=tokenizer,
    #                          get_data_from_file=tfkit.tag.get_data_from_file,
    #                          preprocessing_data=tfkit.tag.preprocessing_data,
    #                          input_arg={'maxlen': maxlen}):
    #         print(i)
    #         self.assertTrue(len(i['input']) == maxlen)
    #         self.assertTrue(len(i['target']) == maxlen)
    #
    #     tokenizer = AutoTokenizer.from_pretrained('voidful/albert_chinese_tiny')
    #     feature = tfkit.tag.preprocessing_data(tokenizer, ["B_Thing", "I_Thing", "O"], "狼 煙 逝 去 ， 幽 夢 醒 來 。",
    #                                            target="O O O O O O O O O O", maxlen=5, separator=" ",
    #                                            handle_exceed='slide')
    #     self.assertEqual(len(feature[0]['target']), 5)
    #
    #     print("start_slice")
    #     feature = tfkit.tag.preprocessing_data(tokenizer, ["B_Thing", "I_Thing", "O"], "狼 煙 逝 去 ， 幽 夢 醒 來 。",
    #                                            target="O O O O O O O O O O", maxlen=5, separator=" ",
    #                                            handle_exceed='start_slice')
    #     self.assertEqual(feature[0]['pos'], [0, 4])
    #
    #     print("end_slice")
    #     feature = tfkit.tag.preprocessing_data(tokenizer, ["B_Thing", "I_Thing", "O"], "狼 煙 逝 去 ， 幽 夢 醒 來 。",
    #                                            target="O O O O O O O O O O", maxlen=5, separator=" ",
    #                                            handle_exceed='end_slice')
    #     self.assertEqual(feature[0]['pos'], [6, 10])
    #
    # def testOnce(self):
    #     tokenizer = AutoTokenizer.from_pretrained('voidful/albert_chinese_tiny')
    #     for i in tfkit.once.get_data_from_file(GEN_DATASET):
    #         print(i)
    #     maxlen = 512
    #     for i in TFKitDataset(GEN_DATASET,
    #                          tokenizer=tokenizer,
    #                          get_data_from_file=tfkit.task.once.get_data_from_file,
    #                          preprocessing_data=tfkit.task.once.preprocessing_data,
    #                          input_arg={'maxlen': maxlen}):
    #         print(tokenizer.convert_tokens_to_string(tokenizer.convert_ids_to_tokens(i['input'])))
    #         print(tokenizer.convert_tokens_to_string(tokenizer.convert_ids_to_tokens(i['target'])))
    #         print(i)
    #         self.assertEqual(maxlen, len(i['input']))
    #         self.assertEqual(maxlen, len(i['target']))
    #
    # def testOnceCTC(self):
    #     tokenizer = AutoTokenizer.from_pretrained('voidful/albert_chinese_tiny')
    #     for i in tfkit.oncectc.get_data_from_file(GEN_DATASET):
    #         print(i)
    #     maxlen = 100
    #     for i in TFKitDataset(GEN_DATASET,
    #                          tokenizer=tokenizer,
    #                          get_data_from_file=tfkit.task.oncectc.get_data_from_file,
    #                          preprocessing_data=tfkit.task.oncectc.preprocessing_data,
    #                          input_arg={'maxlen': maxlen}):
    #         print(tokenizer.convert_tokens_to_string(tokenizer.convert_ids_to_tokens(i['input'])))
    #         print(tokenizer.convert_tokens_to_string(tokenizer.convert_ids_to_tokens(i['target_once'])))
    #         print(tokenizer.convert_tokens_to_string(tokenizer.convert_ids_to_tokens(i['target'])))
    #         print(i)
    #         print(len(i['target']))
    #         self.assertEqual(maxlen, len(i['input']))
    #         self.assertEqual(maxlen, len(i['target']))
    #
    # def testCLM(self):
    #     tokenizer = AutoTokenizer.from_pretrained('voidful/albert_chinese_tiny')
    #
    #     for i in tfkit.clm.get_data_from_file(GEN_DATASET):
    #         print(i)
    #
    #     maxlen = 100
    #     for likelihood in ['none', 'neg', 'pos', 'both']:
    #         print(likelihood)
    #         for i in TFKitDataset(GEN_DATASET,
    #                              tokenizer=tokenizer,
    #                              get_data_from_file=tfkit.task.clm.get_data_from_file,
    #                              preprocessing_data=tfkit.task.clm.preprocessing_data,
    #                              input_arg={'maxlen': maxlen, 'likelihood': likelihood}):
    #             start_pos = i['start']
    #             print(likelihood, i)
    #             # self.assertTrue(tokenizer.mask_token_id == i['input'][start_pos])
    #             if 'neg' in likelihood and i['target'][start_pos] == -1:
    #                 self.assertTrue(i['ntarget'][start_pos] != -1)
    #             else:
    #                 self.assertTrue(i['target'][start_pos] != -1)
    #             self.assertEqual(maxlen, len(i['input']))
    #             self.assertEqual(maxlen, len(i['target']))
    #
    # def testOnebyone(self):
    #     tokenizer = AutoTokenizer.from_pretrained('voidful/albert_chinese_tiny')
    #
    #     maxlen = 10
    #
    #     for i in tfkit.onebyone.get_data_from_file(GEN_DATASET):
    #         print(i)
    #
    #     maxlen = 100
    #     for likelihood in ['none', 'neg', 'pos', 'both']:
    #         print(likelihood)
    #         for i in TFKitDataset(GEN_DATASET,
    #                              tokenizer=tokenizer,
    #                              get_data_from_file=tfkit.task.onebyone.get_data_from_file,
    #                              preprocessing_data=tfkit.task.onebyone.preprocessing_data,
    #                              input_arg={'maxlen': maxlen, 'likelihood': likelihood}):
    #             start_pos = i['start']
    #             print(likelihood, i)
    #             self.assertTrue(tokenizer.mask_token_id == i['input'][start_pos])
    #             if 'neg' in likelihood and i['target'][start_pos] == -1:
    #                 self.assertTrue(i['ntarget'][start_pos] != -1)
    #             else:
    #                 self.assertTrue(i['target'][start_pos] != -1)
    #
    #             self.assertEqual(maxlen, len(i['input']))
    #             self.assertEqual(maxlen, len(i['target']))
    #
    # def testGPT(self):
    #     tokenizer = AutoTokenizer.from_pretrained('openai-gpt')
    #     maxlen = 512
    #     for likelihood in ['none', 'neg', 'pos', 'both']:
    #         output = []
    #         for i in TFKitDataset(GEN_DATASET,
    #                              tokenizer=tokenizer,
    #                              get_data_from_file=tfkit.task.onebyone.get_data_from_file,
    #                              preprocessing_data=tfkit.task.onebyone.preprocessing_data,
    #                              input_arg={'maxlen': maxlen, 'likelihood': likelihood}):
    #             start_pos = i['start']
    #             output.extend(tokenizer.convert_ids_to_tokens([i['target'][start_pos]]))
    #             print(output)
    #         print(tokenizer.convert_tokens_to_string(output) + "\n")
    #
    # def testQA(self):
    #     tokenizer = AutoTokenizer.from_pretrained('voidful/albert_chinese_small')
    #     for i in tfkit.qa.get_data_from_file(QA_DATASET):
    #         print(i)
    #     for i in TFKitDataset(QA_DATASET,
    #                          tokenizer=tokenizer,
    #                          get_data_from_file=tfkit.task.qa.get_data_from_file,
    #                          preprocessing_data=tfkit.task.qa.preprocessing_data):
    #         print(i)
    #         self.assertTrue(len(i['input']) <= 512)
    #         self.assertTrue(len(i['target']) == 2)
    #     for i in TFKitDataset(QA_DATASET,
    #                          tokenizer=tokenizer,
    #                          get_data_from_file=tfkit.task.qa.get_data_from_file,
    #                          preprocessing_data=tfkit.task.qa.preprocessing_data):
    #         self.assertTrue(len(i['input']) <= 512)
    #         self.assertTrue(len(i['target']) == 2)
    #
    # def testLen(self):
    #     tokenizer = AutoTokenizer.from_pretrained('voidful/albert_chinese_small')
    #     ds = TFKitDataset(QA_DATASET,
    #                      tokenizer=tokenizer,
    #                      get_data_from_file=tfkit.task.qa.get_data_from_file,
    #                      preprocessing_data=tfkit.task.qa.preprocessing_data)
    #     old_len = ds.__len__()
    #     ds.increase_with_sampling(2)
    #     self.assertEqual(old_len, ds.__len__())
    #
    #     ds = TFKitDataset(QA_DATASET,
    #                      tokenizer=tokenizer,
    #                      get_data_from_file=tfkit.task.qa.get_data_from_file,
    #                      preprocessing_data=tfkit.task.qa.preprocessing_data)
    #     print("before increase_with_sampling", ds.__len__())
    #     ds.increase_with_sampling(50)
    #     print("after increase_with_sampling", ds.__len__())
    #     self.assertTrue(ds.__len__() == 50)
    #
    # def testSeq2seqWithPrev(self):
    #     tokenizer = AutoTokenizer.from_pretrained('facebook/bart-base')
    #     maxlen = 10
    #     for i in tfkit.seq2seq.get_data_from_file(GEN_DATASET):
    #         print("data", i)
    #
    # def testSeq2seqBT(self):
    #     tokenizer = AutoTokenizer.from_pretrained('facebook/bart-base')
    #
    #     for i in tfkit.seq2seqbt.get_data_from_file(GEN_DATASET):
    #         print("data", i)
    #
    #     maxlen = 10



import unittest

import tfkit
from tfkit.utility.dataloader import pad_batch
from torch import Tensor
from transformers import BertTokenizer, AutoModel, AutoTokenizer


class TestModel(unittest.TestCase):


    def testGenerationModel(self):
        input = "See you next time"
        maxlen = 32
        tokenizer = AutoTokenizer.from_pretrained('sshleifer/bart-tiny-random')
        pretrained = AutoModel.from_pretrained('sshleifer/bart-tiny-random')
        # tfkit.task.seq2seq, tfkit.task.once, tfkit.task.oncectc, tfkit.task.clm
        for gmodel in [tfkit.task.seq2seq, tfkit.task.once, tfkit.task.oncectc, tfkit.task.clm]:
            print(str(gmodel))
            model = gmodel.Model(tokenizer, pretrained, maxlen=maxlen)
            preprocessor = gmodel.Preprocessor(tokenizer, maxlen=maxlen, handle_exceed='start_slice', reserved_len=0)
            for preprocessed_item in preprocessor.preprocess(
                    {'task': 'taskA', 'input': input}):
                print("preprocessed_item", preprocessed_item)
                feature = preprocessor.postprocess(preprocessed_item, tokenizer, maxlen=maxlen)
                feature = preprocessor.postprocess_batch(feature)
                print(model(feature, eval=True))
                self.assertTrue(isinstance(model(feature, eval=True), dict))
                model_dict = model(feature, eval=True)
                self.assertTrue('max_item' in model_dict)

            # greedy
            print("greedy")
            result, detail = model.predict(input=input)
            print(result, model_dict)
            self.assertTrue(len(result) == 1)
            self.assertTrue(isinstance(result, list))
            self.assertTrue(isinstance(detail, dict))

            # TopK
            result, detail = model.predict(input=input, decodenum=3, mode='topK', topK=3, filtersim=False)
            print("topK", result)
            self.assertTrue(len(result) == 3)
            self.assertTrue(isinstance(result, list))
            self.assertTrue(isinstance(detail, dict))

            # beamsearch
            result, detail = model.predict(input=input, decodenum=3)
            print("beamsaerch", len(result), result, model_dict)
            self.assertTrue(len(result) == 3)
            self.assertTrue(isinstance(result, list))
            self.assertTrue(isinstance(detail, dict))

            # TopP
            result, detail = model.predict(input=input, decodenum=3, mode='topP', topP=0.8)
            print("TopP", len(result), result, model_dict)
            self.assertTrue(len(result) == 3)
            self.assertTrue(isinstance(result, list))
            self.assertTrue(isinstance(detail, dict))

            # test exceed 512
            result, model_dict = model.predict(input="T " * 540)
            print("test exceed 512", len(result), result, model_dict)
            self.assertTrue(isinstance(result, list))
            self.assertTrue(isinstance(detail, dict))
            print("exceed max len", result)

            result, model_dict = model.predict(input="T " * 550, reserved_len=10)
            print(result)
            self.assertTrue(isinstance(result, list))
            self.assertTrue(isinstance(detail, dict))
            print("exceed max len with reserved len:", result)

    # def testClas(self):
    #     input = "One hundred thirty-four patients suspected of having pancreas cancer successfully underwent gray scale ultrasound examination of the pancreas ."
    #     target = "a"
    #     tokenizer = BertTokenizer.from_pretrained('voidful/albert_chinese_tiny')
    #     pretrained = AutoModel.from_pretrained('voidful/albert_chinese_tiny')
    #     maxlen = 512
    #     model = tfkit.task.clas.Model(tokenizer, pretrained, tasks_detail={"taskA": ["a", "b"]})
    #     preprocessor = tfkit.task.clas.Preprocessor(tokenizer, maxlen=maxlen, handle_exceed='start_slice',
    #                                                 reserved_len=0)
    #     for preprocessed_item in preprocessor.preprocess(
    #             {'task': 'taskA', 'input': input, 'target': target, 'task_dict': {"taskA": ["a", "b"]}}):
    #         feature = preprocessor.postprocess(preprocessed_item, tokenizer, maxlen=maxlen)
    #         for k, v in feature.items():
    #             feature[k] = [v, v]
    #         print(feature)
    #         # test train
    #         print(model(feature))
    #         self.assertTrue(isinstance(model(feature), Tensor))
    #         # test eval
    #         print(model(feature, eval=True))
    #         model_dict = model(feature, eval=True)
    #         print(model_dict)
    #
    #     # test predict
    #     tok_label = model.predict(task="taskA", input=input)
    #     self.assertTrue(len(tok_label) == 2)
    #     # test predict with top k 2
    #     top_k_label, top_k_dict = model.predict(task="taskA", input=input, topK=2)
    #     print("test predict with top k 2, ", top_k_label, top_k_dict)
    #     self.assertTrue(len(top_k_label) == 2)
    #
    #     # test exceed 512
    #     for merge_strategy in ['entropy', 'count', 'prob']:
    #         result, model_dict = model.predict(task="taskA", input=" ".join([str(i) for i in range(2000)]),
    #                                            merge_strategy=merge_strategy, topK=2)
    #         print(result, len(model_dict), model_dict)
    #         self.assertTrue(isinstance(result, list))
    #         self.assertTrue(len(result) == 2)

    # def testQA(self):
    #     input = "梵 語 在 社 交 中 口 頭 使 用 , 並 且 在 早 期 古 典 梵 語 文 獻 的 發 展 中 維 持 口 頭 傳 統 。 在 印 度 , 書 寫 形 式 是 當 梵 語 發 展 成 俗 語 之 後 才 出 現 的 ; 在 書 寫 梵 語 的 時 候 , 書 寫 系 統 的 選 擇 受 抄 寫 者 所 處 地 域 的 影 響 。 同 樣 的 , 所 有 南 亞 的 主 要 書 寫 系 統 事 實 上 都 用 於 梵 語 文 稿 的 抄 寫 。 自 1 9 世 紀 晚 期 , 天 城 文 被 定 為 梵 語 的 標 準 書 寫 系 統 , 十 分 可 能 的 原 因 是 歐 洲 人 有 用 這 種 文 字 印 刷 梵 語 文 本 的 習 慣 。 最 早 的 已 知 梵 語 碑 刻 可 確 定 為 公 元 前 一 世 紀 。 它 們 採 用 了 最 初 用 於 俗 語 而 非 梵 語 的 婆 羅 米 文 。 第 一 個 書 寫 梵 語 的 證 據 , 出 現 在 晚 於 它 的 俗 語 的 書 寫 證 據 之 後 的 幾 個 世 紀 , 這 被 描 述 為 一 種 悖 論 。 在 梵 語 被 書 寫 下 來 的 時 候 , 它 首 先 用 於 行 政 、 文 學 或 科 學 類 的 文 本 。 宗 教 文 本 口 頭 傳 承 , 在 相 當 晚 的 時 候 才 「 不 情 願 」 地 被 書 寫 下 來 。 [Question] 最 初 梵 語 以 什 麼 書 寫 系 統 被 記 錄 下 來 ?"
    #     target = [201, 205]
    #     tokenizer = BertTokenizer.from_pretrained('voidful/albert_chinese_tiny')
    #     pretrained = AutoModel.from_pretrained('voidful/albert_chinese_tiny')
    #     model = tfkit.task.qa.Model(tokenizer, pretrained, maxlen=512)
    #
    #     proc = tfkit.task.qa.Preprocessor(tokenizer, maxlen=512, handle_exceed='start_slice',
    #                                       reserved_len=0)
    #     for items in proc.preprocess({"input": input}):
    #         raw_input = items['raw_input']
    #         feature = proc.postprocess(items, tokenizer, 512)
    #         for k, v in feature.items():
    #             feature[k] = [v]
    #
    #         # test train
    #         print(model(feature))
    #         self.assertTrue(isinstance(model(feature), Tensor))
    #         # test eval
    #         print(model(feature, eval=True))
    #         model_dict = model(feature, eval=True)
    #         self.assertTrue('label_prob_all' in model_dict)
    #         self.assertTrue('label_map' in model_dict)
    #
    #     # test predict
    #     result, model_dict = model.predict(input=input)
    #     print("model_dict", model_dict, input, result)
    #     self.assertTrue('label_prob_all' in model_dict[0])
    #     self.assertTrue('label_map' in model_dict[0])
    #     self.assertTrue(len(result) == 1)
    #
    #     # # test eval top k = 2
    #     # top_k_label, top_k_dict = task.predict(input=input, topK=2)
    #     # print("top_k_label", top_k_label)
    #     # self.assertTrue(len(top_k_label) == 2)
    #
    #     # test exceed 512
    #     for merge_strategy in ['entropy', 'count', 'prob']:
    #         result, model_dict = model.predict(input=" ".join([str(i) for i in range(550)]),
    #                                            handle_exceed='start_slice',
    #                                            merge_strategy=merge_strategy)
    #         print(result, len(model_dict))
    #         self.assertTrue(isinstance(result, list))
    #
    # def testTag(self):
    #     tokenizer = BertTokenizer.from_pretrained('voidful/albert_chinese_small')
    #     pretrained = AutoModel.from_pretrained('voidful/albert_chinese_small')
    #
    #     input = "在 歐 洲 , 梵 語 的 學 術 研 究 , 由 德 國 學 者 陸 特 和 漢 斯 雷 頓 開 創 。 後 來 威 廉 · 瓊 斯 發 現 印 歐 語 系 , 也 要 歸 功 於 對 梵 語 的 研 究 。 此 外 , 梵 語 研 究 , 也 對 西 方 文 字 學 及 歷 史 語 言 學 的 發 展 , 貢 獻 不 少 。 1 7 8 6 年 2 月 2 日 , 亞 洲 協 會 在 加 爾 各 答 舉 行 。 [SEP] 陸 特 和 漢 斯 雷 頓 開 創 了 哪 一 地 區 對 梵 語 的 學 術 研 究 FOB ?"
    #     target = "O A A O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O O"
    #     label = ["O", "A"]
    #
    #     model = tfkit.task.tag.Model(tokenizer=tokenizer, pretrained=pretrained, tasks_detail={"default": label})
    #
    #     # test exceed 512
    #     for merge_strategy in ['count']:
    #         result, model_dict = model.predict(
    #             input="""
    #                   Rundfadsfdsfsfning 明朝（1368年1月23日－1644年4月25日[註 1]）是中國歷史上最後一個由漢族建立的大一統王朝，歷經十二世、十六位皇帝，國祚二百七十六年[參 4]。\n\n元朝末年政治腐敗，種族紛爭，天災不斷，民不聊生，民變暴動屢禁不止，平民朱元璋加入紅巾軍並在其中乘勢崛起，跟隨佔據濠州的郭子興。郭子興死後，朱元璋被當時反抗軍擁立的小明王韓林兒封為左副元帥，並率部眾先後攻占滁州、和州等地，並最終攻佔集慶（今江蘇南京），採取朱升所建議的「高築牆，廣積糧，緩稱王」的政策，以鞏固根據地，讓士兵屯田積糧減少百姓負擔，以示自己為仁義之師而避免受敵。1364年，朱元璋稱吳王，建立西吳政權。1368年，在掃滅陳友諒、張士誠和方國珍等群雄勢力後，朱元璋於當年農曆正月初四日登基稱帝，立國號為大明[參 5]，定都應天府（今南京市），其轄區稱為京師，由因皇室姓朱，故又稱朱明，之後以「驅逐胡虜，恢復中華」[參 6]為號召北伐中原[參 7][參 8]，並收回了燕雲十六州[參 9]，結束蒙元在中國漢地的統治，統一天下。\n\n明初天下大定，經過朱元璋的休養生息，社會經濟得以恢復和發展，國力迅速恢復，史稱洪武之治。朱元璋去世後，其孫朱允炆即位，但其在靖難之役中敗於駐守燕京的朱元璋第四子朱棣，也自此失蹤。朱棣登基後遷都至順天府（今北京市），將北平布政司升為京師，原京師改稱南京[參 3]。成祖朱棣時期，開疆拓土，又派遣鄭和七下西洋，此後許多漢人遠赴海外，國勢達到頂峰，史稱永樂盛世。其後的仁宗和宣宗時期國家仍處於興盛時期，史稱仁宣之治[參 10]。英宗和代宗時期，遭遇土木之變，國力中衰，經于謙等人抗敵，最終解除國家危機。憲宗和孝宗相繼與民休息，孝宗則力行節儉，減免稅賦，百姓安居樂業，史稱弘治中興[參 11]。武宗時期爆發了南巡之爭和寧王之亂。世宗即位初，引發大禮議之爭，他清除宦官和權臣勢力後總攬朝綱，實現嘉靖中興，並於屯門海戰與西草灣之戰中擊退葡萄牙殖民侵略，任用胡宗憲和俞大猷等將領平定東南沿海的倭患。世宗駕崩後經過隆慶新政國力得到恢復，神宗前期任用張居正，推行萬曆新政，國家收入大增，商品經濟空前繁榮、科學巨匠迭出、社會風尚呈現出活潑開放的新鮮氣息，史稱萬曆中興[參 12]。後經過萬曆三大征平定內憂外患，粉碎豐臣秀吉攻占朝鮮進而入明的計劃，然而因為國本之爭，皇帝逐漸疏於朝政，史稱萬曆怠政，同時東林黨爭也帶來了明中期的政治混亂。\n\n萬曆一朝成為明朝由盛轉衰的轉折期[參 13]。光宗繼位不久因紅丸案暴斃，熹宗繼承大統改元天啟，天啟年間魏忠賢閹黨禍亂朝綱，至明思宗即位後剷除閹黨，但閹黨倒臺後，黨爭又起，政治腐敗以及連年天災[註 2][註 3]，導致國力衰退，最終爆發大規模民變。1644年4月25日（舊曆三月十九），李自成所建立的大順軍攻破北京，思宗自縊於煤山，是為甲申之變。隨後吳三桂倒戈相向，滿族建立的滿清入主中原。明朝宗室於江南地區相繼成立南明諸政權，而原本反明的流寇在李自成等領袖死後亦加入南明陣營，這些政權被清朝統治者先後以「為君父報仇」為名各個殲滅，1662年，明朝宗室最後政權被剷除，永曆帝被俘後被殺，滿清又陸續擊敗各地反抗軍，以及攻取台灣、澎湖，1683年，奉大明為正朔的明鄭向清朝投降，漢族抗爭勢力方為清朝所消滅。[參 16]。\n\n明代的核心領土囊括漢地[註 4]，東北到外興安嶺及黑龍江流域[參 19]，後縮為遼河流域；初年北達戈壁沙漠一帶，後改為今長城；西北至新疆哈密，後改為嘉峪關；西南臨孟加拉灣[註 5]，後折回約今雲南境；曾經在今中國東北、新疆東部及西藏等地設有羈縻機構[參 21]。不過，明朝是否實際統治了西藏國際上尚存在有一定的爭議[註 6]。明成祖時期曾短暫征服及統治安南[參 22]，永樂二十二年（1424年），明朝國土面積達到極盛，在東南亞設置舊港宣慰司[註 7]等行政機構，加強對東南洋一帶的管理[參 23][參 24]。\n\n明代商品經濟繁榮，出現商業集鎮，而手工業及文化藝術呈現世俗化趨勢[參 25]。根據《明實錄》所載的人口峰值於成化十五年（1479年）達七千餘萬人[參 26]，不過許多學者考慮到當時存在大量隱匿戶口，故認為明朝人口峰值實際上逾億[參 27]，還有學者認為晚明人口峰值接近2億[註 8]。這一時期，其GDP總量所占的世界比例在中國古代史上也是最高的，1600年明朝GDP總量為960億美元，占世界經濟總量的29.2%，晚明中國人均GDP在600美元[註 9]。\n\n明朝政治則是權力趨於集中，明太祖在誅殺胡惟庸後廢除傳統的丞相制，六部直接對皇帝負責，後來設置內閣；地方上由承宣布政使司、提刑按察使司、都指揮使司分掌權力，加強地方管理。仁宗、宣宗之後，文官治國的思想逐漸濃厚，行政權向內閣和六部轉移。同時還設有都察院等監察機構，為加強對全國臣民的監視，明太祖設立特務機構錦衣衛，明成祖設立東廠，明憲宗時再設西廠（後取消），明武宗又設內行廠（後取消），合稱「廠衛」。但明朝皇帝並非完全獨斷獨行，有許多事還必須經過經廷推、廷議、廷鞫程序，同時，能將原旨退還的給事中亦可對皇權形成制衡。[參 33]到了後期皇帝出現了怠政，宦官行使大權的陋習[參 3]，儘管決策權始終集中在皇帝手中，然而政務大部分已經由內閣處理，此外，到了明代中晚期文官集團的集體意見足以與皇帝抗衡，在遇到事情決斷兩相僵持不下時，也容易產生一種類似於「憲政危機（英語：Constitutional crisis）」的情況，因此「名義上他是天子，實際上他受制於廷臣。」[參 34]但明朝皇權受制於廷臣主要是基於道德上而非法理上，因為明朝當時風氣普遍注重名節，受儒家教育的皇帝通常不願被冠以「昏君」之名。但雖然皇權受制衡，皇帝仍可任意動用皇權，例如明世宗「大禮議」事件最後以廷杖朝臣多人的方式結束[參 35]，明神宗在國本之爭失利後也以長期拒絕參與政事向朝臣們示威[1][2][3]。\n\n有學者認為明代是繼漢唐之後的黃金時期，也被稱為最後一個可以和漢唐媲美的盛世[參 36]。清代張廷玉等修的官修《明史》評價明朝為「治隆唐宋」[註 10]、「遠邁漢唐」[參 37]。
    #                   """,
    #             merge_strategy=merge_strategy, start_contain="B_",
    #             end_contain="I_")
    #         print(result)
    #         self.assertTrue(isinstance(result, list))

        # proc = tfkit.task.tag.Preprocessor(tokenizer, maxlen=512, handle_exceed='start_slice',
        #                                   reserved_len=0)
        # for items in proc.prepare_data({"input": input}):
        #     raw_input = items['raw_input']
        #     feature = proc.postprocess(items, tokenizer, 512)
        #     for k, v in feature.items():
        #         feature[k] = [v]
        #     self.assertTrue(isinstance(model(feature), Tensor))
        #     print(model(feature))
        #     # test eval
        #     model_dict = model(feature, eval=True)
        #     self.assertTrue('label_prob_all' in model_dict)
        #     self.assertTrue('label_map' in model_dict)
        #     self.assertEqual(len(model_dict['label_map']), len(input.split(" ")))
        #
        # # test predict
        # result, model_dict = model.predict(input=input, start_contain="A", end_contain="A")
        # self.assertTrue('label_prob_all' in model_dict[0])
        # self.assertTrue('label_map' in model_dict[0])
        # print("result", result, len(result))
        # self.assertTrue(isinstance(result, list))
        #
        # # test exceed 512
        # for merge_strategy in ['minentropy', 'maxcount', 'maxprob']:
        #     result, model_dict = model.predict(input=" ".join([str(i) for i in range(1000)]),
        #                                        merge_strategy=merge_strategy, start_contain="A", end_contain="A")
        #     print(result)
        #     self.assertTrue(isinstance(result, list))



import torch
from sklearn.preprocessing import MultiLabelBinarizer
from tfkit.utility import tok
from tfkit.utility.datafile import get_multiclas_data_from_file
from tfkit.utility.preprocess import GeneralNLPPreprocessor


class Preprocessor(GeneralNLPPreprocessor):

    def read_file_to_data(self, path):
        return get_multiclas_data_from_file(path)

    def preprocess_component_convert_to_id(self, item, **param_dict):
        item['input'] = self.tokenizer.convert_tokens_to_ids(item['input'])
        yield item

    def postprocess(self, item, tokenizer, maxlen, **kwargs):
        tinput, task = item['input'], item['task']
        row_dict = {'task': list(task.encode("utf-8"))}
        tokenized_input_id = [tok.tok_begin_id(tokenizer)] + tinput + [tok.tok_sep_id(tokenizer)]
        mask_id = [1] * len(tokenized_input_id)
        tokenized_input_id.extend([tokenizer.pad_token_id] * (maxlen - len(tokenized_input_id)))
        mask_id.extend([-1] * (maxlen - len(mask_id)))
        row_dict['input'] = tokenized_input_id
        row_dict['mask'] = mask_id
        row_dict['target'] = [-1]
        if 'target' in item:
            target = item['target']
            if 'multi_label' in task:
                mlb = MultiLabelBinarizer(classes=item['task_dict'][task])
                tar = mlb.fit_transform(target)
                tokenize_label = tar
            else:
                tokenize_label = [item['task_dict'][task].index(target[0])]
            row_dict['target'] = tokenize_label
        return {key: torch.tensor(value) for key, value in row_dict.items()}
